# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import six
from .base import Deserializer
from armet import media_types


if six.PY3:
    from urllib.parse import parse_qsl

else:
    from urlparse import parse_qsl


class URLDeserializer(Deserializer):

    media_types = media_types.URL

    def deserialize(self, request=None, text=None, encoding='utf8'):

        if text is None:
            # Read in the text from the request.
            text = request.read()

        # Ensure we don't attempt to deserialize nothing.
        if text is None:
            raise ValueError

        try:
            # Attempt to desserialize the URL using the
            # URL decoder.
            data = {}
            for name, value in parse_qsl(text, keep_blank_values=True):
                # Ensure values are properly decoded if neccessary.
                if isinstance(value, six.binary_type):
                    value = value.decode(encoding)

                if isinstance(name, six.binary_type):
                    name = name.decode(encoding)

                # Initialize the array.
                if name not in data:
                    data[name] = []

                # Append the data value.
                data[name].append(value)

            return data

        except AttributeError:
            # Something went wront internally; bad input.
            raise ValueError



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import httplib2
import unittest
import socket
import errno
import json
import base64
import six
from . import http

__all__ = [
    'http',
    'is_available',
    'skipUnlessAvailable',
    'Client',
    'ResourceTestCase'
]


def is_available(host='localhost', port=5000):
    try:
        httplib2.Http().request('http://{}:{}/'.format(host, port))
        return True

    except socket.error as ex:
        return ex.errno != errno.ECONNREFUSED


def skipUnlessAvailable(host='localhost', port=5000):
    if not is_available(host, port):
        raise unittest.SkipTest(
            'server is not available at {}:{}'.format(host, port))


class Client:

    def __init__(self, host='localhost', port=5000):
        # Set the default host and port.
        self.host = host
        self.port = port

        # Setup the HTTP/1.1 connection.
        self.connection = httplib2.Http()
        self.connection.follow_redirects = False

    def request(self, path='/', method='GET', body='', headers=None,
                url=None, username=None, password=None):
        if url is None:
            # Construct the URI.
            url = 'http://{}:{}{}'.format(self.host, self.port, path)

        # Default headers.
        if headers is None:
            headers = {}

        # Construct authorization if needed.
        if username or password:
            creds = '{}:{}'.format(username or '', password or '')
            creds = base64.b64encode(creds.encode('utf8'))
            headers['authorization'] = 'basic ' + creds.decode('utf8')

        # Serialize the body if neccessary.
        # TODO: Support more than JSON.
        if not isinstance(body, six.string_types):
            body = json.dumps(body)
            headers['content-type'] = 'application/json'

        # Perform the response. Remember this does not go anywhere; it is
        # caught by wsgi-intercept and mocked into Armet.
        return self.connection.request(
            url, method,
            body=body,
            headers=headers)

    def options(self, *args, **kwargs):
        kwargs.setdefault('method', 'OPTIONS')
        return self.request(*args, **kwargs)

    def head(self, *args, **kwargs):
        kwargs.setdefault('method', 'HEAD')
        return self.request(*args, **kwargs)

    def get(self, *args, **kwargs):
        return self.request(*args, **kwargs)

    def post(self, *args, **kwargs):
        kwargs.setdefault('method', 'POST')
        return self.request(*args, **kwargs)

    def put(self, *args, **kwargs):
        kwargs.setdefault('method', 'PUT')
        return self.request(*args, **kwargs)

    def delete(self, *args, **kwargs):
        kwargs.setdefault('method', 'DELETE')
        return self.request(*args, **kwargs)


class ResourceTestCase(unittest.TestCase):

    host = 'localhost'

    port = 5000

    def setUp(self):
        # Skip this test module unless the server is available.
        skipUnlessAvailable(self.host, self.port)

        # Initialize the test client.
        self.client = Client(self.host, self.port)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import six
import ujson as json
from .base import Deserializer
from armet import media_types


class JSONDeserializer(Deserializer):

    media_types = media_types.JSON

    def deserialize(self, request=None, text=None, encoding='utf8'):
        if text is None:
            # Read in the text from the request.
            text = request.read()

        # Ensure the text is decoded.
        if isinstance(text, six.binary_type):
            text = text.decode(encoding)

        try:
            # Attempt to deserialize the text.
            return json.loads(text)

        except TypeError:
            # Failed; possibly null.
            raise ValueError



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import six
from .attribute import Attribute
import datetime
from armet import exceptions


try:
    # Attempt to import and make use of date/time libraries.
    # TODO: Make note of this optional dep somewhere.
    from dateutil.parser import parse as parse_datetime
    from time import mktime

except ImportError:
    # No support for date/times.
    parse_datetime = None


class _TemporalAttribute(Attribute):
    """Represents a temporal attribute, such as a date or time.
    """

    def __init__(self, *args, **kwargs):
        # Ensure we have support.
        if parse_datetime is None:
            raise exceptions.ImproperlyConfigured(
                'Use of temporal attributes requires at '
                'least python-dateutil (and optionally parsedatetime).')

        # Continue on.
        super(_TemporalAttribute, self).__init__(*args, **kwargs)

    def prepare(self, value):
        if not value:
            return None

        # Serialize as ISO format.
        return value.isoformat()

    def clean(self, value):
        if not value:
            # Value is nothing; return it.
            return value

        try:
            # Attempt to use the dateutil library to parse.
            return parse_datetime(value, fuzzy=False)

        except (ValueError, AttributeError):
            # Not a strictly formatted date; return nothing.
            pass

        try:
            # Attempt to magic a date out of it.
            # TODO: List this somewhere as an optional dep.
            from parsedatetime import parsedatetime as pdt
            c = pdt.Constants()
            c.BirthdayEpoch = 80  # TODO: Figure out what this is.
            p = pdt.Calendar(c)
            result = p.parse(value)
            if result[1] != 0:
                return datetime.datetime.fromtimestamp(mktime(result[0]))

        except (NameError, ImportError, TypeError):
            # No magical date/time support.
            pass

        # Couldn't figure out what we're dealing with.
        raise ValueError('Invalid date/time or in an invalid format.')


class DateAttribute(_TemporalAttribute, Attribute):

    type = datetime.date

    def clean(self, value):
        value = super(DateAttribute, self).clean(value)
        if value:
            # Constrain to just the date part.
            value = value.date()
        return value


class TimeAttribute(_TemporalAttribute, Attribute):

    type = datetime.time

    def clean(self, value):
        value = super(TimeAttribute, self).clean(value)
        if value:
            # Constrain to just the date part.
            value = value.time()
        return value


class DateTimeAttribute(_TemporalAttribute, Attribute):

    type = datetime.datetime



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import json
from armet import http, serializers, deserializers
from .base import BaseResourceTest


class TestResourcePost(BaseResourceTest):

    def test_post_single(self, connectors):
        data = {'question': 'Is anybody really out there?'}
        body = json.dumps(data)
        response, content = self.client.post(
            path='/api/poll/', body=body,
            headers={'Content-Type': 'application/json'})

        assert response.status == http.client.CREATED

        data = json.loads(content.decode('utf8'))

        assert isinstance(data, dict)
        assert data['question'] == 'Is anybody really out there?'
        assert data['id'] == 101


class TestResourceEcho(BaseResourceTest):

    @classmethod
    def setup_class(cls):
        super(TestResourceEcho, cls).setup_class()

        cls.body = {'x': ['1', '2', '3'], 'y': ['4', '5']}

        cls.serializers = {}
        cls.serializers['json'] = serializers.JSONSerializer()
        cls.serializers['url'] = serializers.URLSerializer()

        cls.deserializers = {}
        cls.deserializers['json'] = deserializers.JSONDeserializer()
        cls.deserializers['url'] = deserializers.URLDeserializer()

    def echo(self, in_format, out_format):
        body = self.serializers[in_format].serialize(self.body)
        response, content = self.client.post(
            path='/api/echo/', body=body,
            headers={
                'Content-Type': self.serializers[in_format].media_types[0],
                'Accept': self.serializers[out_format].media_types[0]})

        assert response.status == http.client.OK

        data = self.deserializers[out_format].deserialize(text=content)

        assert data == self.body

    def test_echo_json_json(self, connectors):
        self.echo('json', 'json')

    def test_echo_json_url(self, connectors):
        self.echo('json', 'url')

    def test_echo_url_url(self, connectors):
        self.echo('url', 'url')

    def test_echo_url_json(self, connectors):
        self.echo('url', 'json')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import sys
import armet
from armet import resources, attributes, exceptions

# Request the generic models module inserted by the test runner.
models = sys.modules['tests.connectors.models']

__all__ = [
    'SimpleResource',
    'SimpleTrailingResource',
    'PollResource',
    'StreamingResource',
    'AsyncResource',
    'AsyncStreamResource',
    'lightweight',
    'lightweight_streaming',
    'lightweight_async',
    'LeftResource',
    'RightResource',
    'echo',
    'cookie',
    'DirectResource',
    'ModelDirectResource',
    'IndirectResource',
    'ModelIndirectResource',
    'TwiceIndirectResource',
    'ModelTwiceIndirectResource',
    'ThriceIndirectResource',
    'ModelThriceIndirectResource',
    'MixinResource',
    'PollExcludeResource',
    'PollUnreadResource',
    'PollUnwriteResource',
    'PollNoNullResource',
    'PollRequiredResource',
    'PollValidResource',
    'PollNamedResource',
]


class SimpleResource(resources.Resource):

    def get(self, request, response):
        # Do nothing and return nothing.
        pass


class SimpleTrailingResource(resources.Resource):

    class Meta:
        trailing_slash = False

    def get(self, request, response):
        # Do nothing and return nothing.
        pass


class StreamingResource(resources.Resource):

    def get(self, request, response):
        response.status = 412
        response['Content-Type'] = 'text/plain'

        yield 'this\n'

        response.write('where\n')
        response.flush()
        response.write('whence\n')

        yield
        yield 'that\n'

        response.write('why\n')

        yield 'and the other'


def spawn(connectors, function):
    import gevent
    gevent.spawn(function)


class AsyncResource(resources.Resource):

    class Meta:
        asynchronous = True

    def get(self, request, response):

        def writer():
            response.status = 412
            response['Content-Type'] = 'text/plain'
            response.write('Hello')
            response.close()

        spawn(self.meta.connectors, writer)


class AsyncStreamResource(resources.Resource):

    class Meta:
        asynchronous = True

    def get(self, request, response):

        def streamer():
            response.status = 412
            response['Content-Type'] = 'text/plain'
            response.write('this\n')
            response.flush()

            response.write('where\n')
            response.flush()

            response.write('whence\n')
            response.write('that\n')
            response.flush()

            response.write('why\n')
            response.write('and the other')
            response.close()

        spawn(self.meta.connectors, streamer)


@armet.resource(methods='GET')
def lightweight(request, response):
    response.status = 412
    response['Content-Type'] = 'text/plain'
    response.write('Hello')


@armet.resource(methods='POST')
def lightweight(request, response):
    response.status = 414
    response['Content-Type'] = 'text/plain'
    response.write('Hello POST')


@armet.resource(methods='GET')
def lightweight_streaming(request, response):
    response.status = 412
    response['Content-Type'] = 'text/plain'

    yield 'this\n'

    response.write('where\n')
    response.flush()
    response.write('whence\n')

    yield
    yield 'that\n'

    response.write('why\n')

    yield 'and the other'


@armet.asynchronous
@armet.resource(methods='GET')
def lightweight_async(request, response):
    def writer():
        response.status = 412
        response['Content-Type'] = 'text/plain'
        response.write('Hello')
        response.close()

    spawn(lightweight_async.meta.connectors, writer)


class PollResource(resources.ModelResource):

    class Meta:
        model = models.Poll

        slug = 'id'

    id = attributes.IntegerAttribute('id')

    question = attributes.TextAttribute('question')

    available = attributes.BooleanAttribute('available')


class LeftResource(resources.Resource):

    class Meta:
        http_allowed_methods = ('HEAD', 'GET', 'OPTIONS',)
        http_allowed_origins = ('http://127.0.0.1:80',)
        http_allowed_headers = ('Content-Type', 'Content-MD5', 'Accept',)

    def get(self):
        # Do nothing.
        pass


class RightResource(resources.Resource):

    class Meta:
        http_allowed_methods = ('HEAD', 'GET', 'DELETE', 'OPTIONS',)
        http_allowed_origins = ('*',)
        http_allowed_headers = ('Content-Type', 'Content-MD5', 'Accept')

    def get(self):
        # Do nothing.
        pass


@armet.resource(methods='POST')
def echo(request, response):
    # Read in the given data in the given format.
    data = request.read(deserialize=True)

    # Write out the parsed data in the requested format.
    response.write(data, serialize=True)


@armet.resource(methods='GET')
def cookie(request, response):
    response['Content-Type'] = 'text/plain'
    response.write(request.cookies['blue'].value)


class DirectResource(resources.Resource):

    def route(self, request, response):
        return super(DirectResource, self).route(request, response)

    def get(self, request, response):
        response.write(b'42')


class ModelDirectResource(resources.ModelResource):

    class Meta:
        model = models.Poll

    id = attributes.IntegerAttribute('id')

    question = attributes.TextAttribute('question')

    def route(self, request, response):
        return super(ModelDirectResource, self).route(request, response)


class IndirectResource(DirectResource):

    def route(self, request, response):
        return super(IndirectResource, self).route(request, response)

    def get(self, request, response):
        response.write(b'84')


class ModelIndirectResource(ModelDirectResource):

    class Meta:
        model = models.Poll

    def read(self):
        return super(ModelIndirectResource, self).read()


class TwiceIndirectResource(IndirectResource):

    def route(self, request, response):
        return super(TwiceIndirectResource, self).route(request, response)

    def get(self, request, response):
        response.write(b'84')


class ModelTwiceIndirectResource(ModelIndirectResource):

    class Meta:
        model = models.Poll

    def read(self):
        return super(ModelTwiceIndirectResource, self).read()


class ThriceIndirectResource(IndirectResource):

    def route(self, request, response):
        return super(ThriceIndirectResource, self).route(request, response)

    def get(self, request, response):
        response.write(b'84')


class ModelThriceIndirectResource(ModelIndirectResource):

    class Meta:
        model = models.Poll

    def read(self):
        return super(ModelThriceIndirectResource, self).read()


class ExtraStuff(object):

    def dispatch(self, *args):
        self.content = b'Hello'
        return super(ExtraStuff, self).dispatch(*args)


class MixinResource(ExtraStuff, IndirectResource):

    content = b'World'

    def route(self, request, response):
        return super(MixinResource, self).route(request, response)

    def get(self, request, response):
        response.write(self.content)


class PollExcludeResource(PollResource):

    question = attributes.TextAttribute('question', include=False)


class PollUnreadResource(PollResource):

    question = attributes.TextAttribute('question', read=False)


class PollUnwriteResource(PollResource):

    question = attributes.TextAttribute('question', write=False)


class PollNoNullResource(PollResource):

    question = attributes.TextAttribute('question', null=False)


class PollRequiredResource(PollResource):

    question = attributes.TextAttribute('question', required=True)


class PollNamedResource(PollResource):

    question = attributes.TextAttribute('question', name='superQuestion')


class PollValidResource(PollResource):

    votes = attributes.IntegerAttribute('votes')

    def clean_votes(self, value):
        assert value > 0, 'Must be greater than 0.'
        assert value < 51, 'Must be less than 51.'
        return value

    def clean_question(self, value):
        errors = []

        if len(value) <= 15:
            errors.append('Must be more than 15 characters.')

        if value.find('?') == -1:
            errors.append('Must have at least one question mark.')

        if errors:
            raise exceptions.ValidationError(*errors)

        return value



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import six
from .base import Serializer
from armet import media_types

# Import urlencode
if six.PY3:
    from urllib.parse import urlencode

else:
    from urllib import urlencode


class URLSerializer(Serializer):

    media_types = media_types.URL

    def serialize(self, obj=None):
        # If we have nothing; serialize as an empty object.
        if obj is None:
            return ''

        try:
            # Attempt to serialize the incoming object using the URL encoder.
            return super(URLSerializer, self).serialize(urlencode(obj, True))

        except TypeError:
            # Raise up our hands; we cannot serialize this.
            raise ValueError



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import six
from .attribute import Attribute
import uuid


class UUIDAttribute(Attribute):

    type = uuid.UUID

    def prepare(self, value):
        # Serialize as the 16-digit hex representation.
        return value.hex if value else value

    def clean(self, value):
        if value is None:
            # Value is nothing; return it.
            return value

        try:
            # Attempt to coerce the UUID.
            return uuid.UUID(value)

        except ValueError:
            raise ValueError(
                'UUID must be of the form: '
                'xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
from .decorators import classproperty, boundmethod
from .functional import cons, compose
from .string import dasherize
from .package import import_module

__all__ = [
    'classproperty',
    'boundmethod',
    'cons',
    'compose',
    'import_module',
    'dasherize',
    'super'
]


class Superclass(object):

    def __init__(self, instance, cls, mro, index):
        self._index = index
        self._instance = instance
        self._cls = cls
        self._mro = mro

    def __getattr__(self, name):
        index = self._index
        while index < len(self._mro):
            cls = self._mro[index]
            if name in cls.__dict__:
                return cls.__dict__[name].__get__(self._instance, self._cls)

            index += 1


def super(current, obj):
    if isinstance(obj, type):
        # Get the MRO for the passed object.
        mro = obj.__mro__

        # Resolve an instance and class object from the object.
        cls, self = obj, None

    else:
        # Get the MRO for the passed object.
        mro = type(obj).__mro__

        # Resolve an instance and class object from the object.
        cls, self = type(obj), obj

    # Resolve the superclass.
    return Superclass(obj, cls, mro, mro.index(current) + 1)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import six
import armet
from armet import utils
from . import options


#! Map of connector class objects in their connector modules.
CONNECTORS = {
    'http': {
        'resource': ('{}.resources', 'Resource',),
        'options': ('{}.resources', 'ResourceOptions',)
    },
    'model': {
        'resource': ('{}.resources', 'ModelResource',),
        'options': ('{}.resources', 'ModelResourceOptions',)
    }
}


#! Set of connector classes.
connector_set = set()


class ResourceBase(type):

    #! Options class to use to expand options.
    options = options.ResourceOptions

    #! Connectors to instantiate and mixin to the inheritance.
    connectors = ['http']

    def mro(self):
        if not self._is_resource(self.__name__, self.__bases__):
            # Not a resource; perform the normal MRO.
            return super(ResourceBase, self).mro()

        # Retrieve the normal MRO.
        mro = super(ResourceBase, self).mro()

        # Remove all connectors from the mro.
        for cls in list(mro):
            if cls in connector_set:
                mro.remove(cls)

        # Attempt to resolve a list of connectors.
        meta = self.__dict__.get('meta')
        if meta:
            cmap = CONNECTORS
            for key, ref in six.iteritems(meta.connectors):
                if key not in self.connectors:
                    continue

                name = cmap[key]['resource'][0].format(ref)
                module = utils.import_module(name)
                if module:
                    klass = getattr(module, cmap[key]['resource'][1], None)
                    if klass:
                        # Found a connector class for this connector.

                        # Find where in the MRO it should go.
                        # It needs to go directly after the resource base
                        # for which it is a connector for.
                        for index, cls in enumerate(mro):
                            if key in getattr(cls, 'connectors', []):
                                break

                        # Let us forever know this is a connector class.
                        connector_set.update(set(klass.__mro__) - set(mro))

                        # Insert the connector class.
                        for cls in klass.__mro__:
                            if cls not in mro:
                                index += 1
                                mro.insert(index, cls)

        # Return the constructed MRO.
        return mro

    @classmethod
    def _is_resource(cls, name, bases):
        if name == 'NewBase':
            # This is a six contrivance; not a real class.
            return False

        if name.startswith('armet.connector:'):
            # This is special mixed connector class; not something
            # to run the metaclass over.
            return False

        for base in bases:
            if base.__name__ == 'NewBase':
                # This is a six contrivance; move along.
                continue

            if isinstance(base, cls):
                # This is some sort of derived resource; good.
                return True

        # This is not derived at all from Resource (eg. is base).
        return False

    @classmethod
    def _gather_metadata(cls, metadata, bases):
        for base in bases:
            if isinstance(base, cls) and hasattr(base, 'Meta'):
                # Append metadata.
                metadata.append(getattr(base, 'Meta'))

                # Recurse.
                cls._gather_metadata(metadata, base.__bases__)

    def __new__(cls, name, bases, attrs):
        if not cls._is_resource(name, bases):
            # This is not an actual resource.
            return super(ResourceBase, cls).__new__(cls, name, bases, attrs)

        # Gather the attributes of all options classes.
        # Start with the base configuration.
        metadata = armet.use().copy()
        values = lambda x: {n: getattr(x, n) for n in dir(x)}

        # Expand the options class with the gathered metadata.
        base_meta = []
        cls._gather_metadata(base_meta, bases)

        # Apply the configuration from each class in the chain.
        for meta in base_meta:
            metadata.update(**values(meta))

        cur_meta = {}
        if attrs.get('Meta'):
            # Apply the configuration from the current class.
            cur_meta = values(attrs['Meta'])
            metadata.update(**cur_meta)

        # Gather and construct the options object.
        meta = attrs['meta'] = cls.options(metadata, name, cur_meta, base_meta)

        # Construct the class object.
        self = super(ResourceBase, cls).__new__(cls, name, bases, attrs)

        # Generate a serializer map that maps media ranges to serializer
        # names.
        self._serializer_map = smap = {}
        for key in self.meta.allowed_serializers:
            serializer = self.meta.serializers[key]
            for media_type in serializer.media_types:
                smap[media_type] = key

        # Generate a deserializer map that maps media ranges to deserializer
        # names.
        self._deserializer_map = dmap = {}
        for key in self.meta.allowed_deserializers:
            deserializer = self.meta.deserializers[key]
            for media_type in deserializer.media_types:
                dmap[media_type] = key

        # Filter the available connectors according to the
        # metaclass restriction set.
        for key in list(meta.connectors.keys()):
            if key not in cls.connectors:
                del meta.connectors[key]

        # Iterate through the available connectors.
        cmap = CONNECTORS
        for key, ref in six.iteritems(meta.connectors):
            options = utils.import_module(cmap[key]['options'][0].format(ref))
            if options:
                options = getattr(options, cmap[key]['options'][1], None)
                if options:
                    # Available options to parse for this connector;
                    # instantiate the options class and apply all
                    # available options.
                    options_instance = options(metadata, name, base_meta)
                    meta.__dict__.update(**options_instance.__dict__)

        # Return the constructed instance.
        return self



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
from .exceptions import ImproperlyConfigured
import six
from . import utils


class route:
    """
    Mounts a resource on the passed URL mount point by invoking the
    `mount` classmethod.

    @example
        @route('/')
        class Resource(resources.Resource):
            # ...
    """

    def __init__(self, *args, **kwargs):
        # Just store the arguments.
        self.args = args
        self.kwargs = kwargs

    def __call__(self, resource):
        # Is the resources a function?
        if callable(resource) and not isinstance(resource, type):
            # Yes; call the resource decorator.
            import armet
            resource = armet.resource(**self.kwargs)(resource)

        # Ensure the resource has a `mount` classmethod.
        if not hasattr(resource, 'mount'):
            raise ImproperlyConfigured(
                'The {} resource doesn\'t have a `mount` method.'.format(
                    resource.meta.name))

        # Hook up the resource at the mount point.
        resource.mount(*self.args)

        # Return the resource
        return resource


def asynchronous(resource):
    """Instructs a decorated resource that it is to be asynchronous.

    An asynchronous resource means that `response.close()` must be called
    explicitly as returning from a method (eg. `get`) does not close
    the connection.

    @note
        This can also be configured by setting `asynchronous` to `True`
        on `<resource>.Meta`. The benefit of the decorator is that this
        can be applied to specific methods as well as the entire class
        body.
    """
    # TODO: Check for gevent support...

    # Flip the asynchronous switch.
    resource.meta.asynchronous = True

    # Return the resource
    return resource


#! Mapping of lightweight resources.
_resources = {}

#! Mapping of functional handlers for the lightweight resources.
_handlers = {}


def resource(**kwargs):
    """Wraps the decorated function in a lightweight resource."""
    def inner(function):
        name = kwargs.pop('name', None)
        if name is None:
            name = utils.dasherize(function.__name__)

        methods = kwargs.pop('methods', None)
        if isinstance(methods, six.string_types):
            # Tuple-ify the method if we got just a string.
            methods = methods,

        # Construct a handler.
        handler = (function, methods)

        if name not in _resources:
            # Initiate the handlers list.
            _handlers[name] = []

            # Construct a light-weight resource using the passed kwargs
            # as the arguments for the meta.
            from armet import resources
            kwargs['name'] = name

            class LightweightResource(resources.Resource):
                Meta = type(str('Meta'), (), kwargs)

                def route(self, request, response):
                    for handler, methods in _handlers[name]:
                        if methods is None or request.method in methods:
                            return handler(request, response)

                    resources.Resource.route(self)

            # Construct and add this resource.
            _resources[name] = LightweightResource

        # Add this to the handlers.
        _handlers[name].append(handler)

        # Return the resource.
        return _resources[name]

    # Return the inner method.
    return inner



# -*- coding: utf-8 -*-
# from __future__ import absolute_import, unicode_literals, division
from os import path

DEBUG = True
TEMPLATE_DEBUG = DEBUG

PROJECT_ROOT = path.abspath(path.dirname(__file__))

ADMINS = (
    # ('Your Name', 'your_email@example.com'),
)

MANAGERS = ADMINS

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
        'USER': '',
        'PASSWORD': '',
        'HOST': '',
        'PORT': '',
    }
}

# Turn off auto slash handling.
APPEND_SLASH = False

# Hosts/domain names that are valid for this site; required if DEBUG is False
# See https://docs.djangoproject.com/en/1.5/ref/settings/#allowed-hosts
ALLOWED_HOSTS = []

# Local time zone for this installation. Choices can be found here:
# http://en.wikipedia.org/wiki/List_of_tz_zones_by_name
# although not all choices may be available on all operating systems.
# In a Windows environment this must be set to your system time zone.
TIME_ZONE = 'America/Chicago'

# Language code for this installation. All choices can be found here:
# http://www.i18nguy.com/unicode/language-identifiers.html
LANGUAGE_CODE = 'en-us'

SITE_ID = 1

# URL configuration.
ROOT_URLCONF = 'tests.connectors.django.urls'

# If you set this to False, Django will make some optimizations so as not
# to load the internationalization machinery.
USE_I18N = True

# If you set this to False, Django will not format dates, numbers and
# calendars according to the current locale.
USE_L10N = True

# If you set this to False, Django will not use timezone-aware datetimes.
USE_TZ = True

# Absolute filesystem path to the directory that will hold user-uploaded files.
# Example: "/var/www/example.com/media/"
MEDIA_ROOT = ''

# URL that handles the media served from MEDIA_ROOT. Make sure to use a
# trailing slash.
# Examples: "http://example.com/media/", "http://media.example.com/"
MEDIA_URL = ''

# Absolute path to the directory static files should be collected to.
# Don't put anything in this directory yourself; store your static files
# in apps' "static/" subdirectories and in STATICFILES_DIRS.
# Example: "/var/www/example.com/static/"
STATIC_ROOT = ''

# URL prefix for static files.
# Example: "http://example.com/static/", "http://static.example.com/"
STATIC_URL = '/static/'

# Additional locations of static files
STATICFILES_DIRS = (
    # Put strings here, like "/home/html/static" or "C:/www/django/static".
    # Always use forward slashes, even on Windows.
    # Don't forget to use absolute paths, not relative paths.
)

# List of finder classes that know how to find static files in
# various locations.
STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
)

# Make this unique, and don't share it with anybody.
SECRET_KEY = '_lhu2gqrh7j0+9aa_*n-fzerhsar+n$tm1nf+6i+f+$abx#$q@'

# List of callables that know how to import templates from various sources.
TEMPLATE_LOADERS = (
    'django.template.loaders.filesystem.Loader',
    'django.template.loaders.app_directories.Loader',
)

MIDDLEWARE_CLASSES = (
    'django.middleware.common.CommonMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    # Uncomment the next line for simple clickjacking protection:
    # 'django.middleware.clickjacking.XFrameOptionsMiddleware',
)

TEMPLATE_DIRS = (
    # Put strings here, like "/home/html/django_templates"
    # Always use forward slashes, even on Windows.
    # Don't forget to use absolute paths, not relative paths.
)

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'tests.connectors.django'
)

# A sample logging configuration. The only tangible logging
# performed by this configuration is to send an email to
# the site admins on every HTTP 500 error when DEBUG=False.
# See http://docs.djangoproject.com/en/dev/topics/logging for
# more details on how to customize your logging configuration.
LOGGING = {
    'version': 1,
    'disable_existing_loggers': True,
    'formatters': {
        'simple': {
            'format': '%(levelname)s %(message)s'
        },
    },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'simple'
        },
    },
    'loggers': {
        'armet': {
            'handlers': ['console'],
            'level': 'DEBUG',
        },
    }
}



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals, division
import importlib


def import_module(name):
    """Attempt to import a module; returns None if unsuccessful."""
    try:
        return importlib.import_module(name)

    except ImportError:
        return None



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import json
from armet import http
from .base import BaseResourceTest


class TestResourcePut(BaseResourceTest):

    def test_put_existing(self, connectors):
        data = {'id': 1, 'question': 'Is anybody really out there?'}
        body = json.dumps(data)
        response, content = self.client.put(
            path='/api/poll/1/', body=body,
            headers={'Content-Type': 'application/json'})

        assert response.status == http.client.OK

        data = json.loads(content.decode('utf8'))

        assert isinstance(data, dict)
        assert data['question'] == 'Is anybody really out there?'
        assert data['id'] == 1



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import collections
import flask
from werkzeug.routing import BaseConverter
from . import http
from armet import utils


class RegexConverter(BaseConverter):
    """Regular expression URL converter for werkzeug / flask.
    """

    def __init__(self, url_map, pattern=r'(.*)'):
        super(RegexConverter, self).__init__(url_map)
        self.regex = pattern


class Resource(object):

    @classmethod
    def view(cls, *args, **kwargs):
        # Initiate the base view request cycle.
        path = kwargs.get('path', '')
        if flask.request.path.endswith('/'):
            # The trailing slash is stripped for fun by werkzeug.
            path += '/'

        # Construct request and response wrappers.
        async = cls.meta.asynchronous
        request = http.Request(path=path, asynchronous=async)
        response = http.Response(asynchronous=async)

        # Defer the execution thread if we're running asynchronously.
        if async:
            # Defer the view to pass of control.
            import gevent
            gevent.spawn(utils.super(Resource, cls).view, request, response)

            # Construct and return the generator response.
            response._handle.response = cls.stream(response, response)
            return response._handle

        # Pass control off to the resource handler.
        result = utils.super(Resource, cls).view(request, response)

        if isinstance(result, collections.Iterator):
            # Construct and return the generator response.
            response._handle.response = result
            return response._handle

        # Configure the response if we received any data.
        if result is not None:
            response._handle.data = result

        # Return the response.
        return response._handle

    @classmethod
    def mount(cls, url, app=None):
        # Generate a name to use to mount this resource.
        name = 'armet' + cls.__module__ + cls.meta.name + url

        # If application is not provided; make use of the app context.
        if app is None:
            app = flask.current_app

        # Prepare the flask application to accept armet resources.
        # The strict slashes setting must be False for our routes.
        strict_slashes = app.url_map.strict_slashes = False

        # Ensure that there is a compliant regex converter available.
        converter = app.url_map.converters['default']
        app.url_map.converters['default'] = RegexConverter

        # Mount this resource.
        pattern = '{}{}<path>'.format(url, cls.meta.name)
        rule = app.url_rule_class(pattern, endpoint=name)
        app.url_map.add(rule)
        app.view_functions[name] = cls.view

        # Restore the flask environment.
        app.url_map.strict_slashes = strict_slashes
        app.url_map.converters['default'] = converter



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division


#! Equality
OPERATOR_EQUAL = 'exact', '=='

#! Case-insensitve equality
OPERATOR_IEQUAL = 'iexact', '='

#! Less than
OPERATOR_LT = 'lt', '<'

#! Less than or equal to
OPERATOR_LTE = 'lte', '<='

#! Greater than
OPERATOR_GT = 'gt', '>'

#! Greater than or equal to
OPERATOR_GTE = 'gte', '>='

#! Regular expression match
OPERATOR_REGEX = 'regex', '*='

#! Null test
OPERATOR_ISNULL = 'isnull', None

#! In-collection test
OPERATOR_IN = 'in', None

#! Operators
OPERATORS = (
    OPERATOR_EQUAL,
    OPERATOR_IEQUAL,
    OPERATOR_LT,
    OPERATOR_LTE,
    OPERATOR_GT,
    OPERATOR_GTE,
    OPERATOR_REGEX,
    OPERATOR_ISNULL,
    OPERATOR_IN)

#! Negation
NEGATION = ('not', '!')

#! Logical
LOGICAL_AND = '&'
LOGICAL_OR = ';'

#! Path separator
SEP_PATH = '.'

#! Value separator
SEP_VALUE = ','

#! Directive initiator
DIRECTIVE = ':'

#! Grouping characters
GROUP_BEGIN = '('
GROUP_END = ')'



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import json
from armet import http
from .base import BaseResourceTest


class TestResourceAttributeProperties(BaseResourceTest):

    def test_not_include(self):
        response, content = self.client.get('/api/poll-exclude/1/')

        assert response.status == http.client.OK

        data = json.loads(content.decode('utf8'))

        assert 'id' in data
        assert 'question' not in data

    def test_not_include_access(self):
        response, content = self.client.get('/api/poll-exclude/1/question/')

        assert response.status == http.client.OK
        assert content.decode('utf8') == '["Are you an innie or an outie?"]'

    def test_not_read(self):
        response, content = self.client.get('/api/poll-unread/1/question/')

        assert response.status == http.client.FORBIDDEN

    def test_not_include_but_create(self):
        data = {'question': 'Is anybody really out there?'}
        body = json.dumps(data)
        response, content = self.client.post(
            path='/api/poll-exclude/', body=body,
            headers={'Content-Type': 'application/json'})

        assert response.status == http.client.CREATED

        data = json.loads(content.decode('utf8'))

        assert 'question' not in data
        assert data['id'] == 101

        response, content = self.client.get('/api/poll/101/')

        assert response.status == http.client.OK

        data = json.loads(content.decode('utf8'))

        assert data['question'] == 'Is anybody really out there?'

    def test_not_include_but_replace(self):
        data = {'id': 1, 'question': 'Is anybody really out there?'}
        body = json.dumps(data)
        response, content = self.client.put(
            path='/api/poll-exclude/1/', body=body,
            headers={'Content-Type': 'application/json'})

        assert response.status == http.client.OK

        data = json.loads(content.decode('utf8'))

        assert 'question' not in data
        assert data['id'] == 1

        response, content = self.client.get('/api/poll/1/')

        assert response.status == http.client.OK

        data = json.loads(content.decode('utf8'))

        assert data['question'] == 'Is anybody really out there?'

    def test_not_write(self):
        data = {'id': 123, 'question': 'gr9uer0gn2w'}
        body = json.dumps(data)
        response, content = self.client.put(
            path='/api/poll-unwrite/1/', body=body,
            headers={'Content-Type': 'application/json'})

        assert response.status == http.client.BAD_REQUEST

        data = json.loads(content.decode('utf8'))

        assert data['question'] == ['Attribute is read-only.']

    def test_not_write_same(self):
        data = {'id': 1, 'question': 'Is anybody really out there?'}
        body = json.dumps(data)
        response, content = self.client.put(
            path='/api/poll-unwrite/1/', body=body,
            headers={'Content-Type': 'application/json'})

        assert response.status == http.client.OK

    def test_no_null(self):
        data = {'question': None}
        body = json.dumps(data)
        response, content = self.client.put(
            path='/api/poll-no-null/1/', body=body,
            headers={'Content-Type': 'application/json'})

        assert response.status == http.client.BAD_REQUEST

        data = json.loads(content.decode('utf8'))

        assert data['question'] == ['Must not be null.']

    def test_required(self):
        data = {}
        body = json.dumps(data)
        response, content = self.client.put(
            path='/api/poll-required/1/', body=body,
            headers={'Content-Type': 'application/json'})

        assert response.status == http.client.BAD_REQUEST

        data = json.loads(content.decode('utf8'))

        assert data['question'] == ['Must be provided.']

    def test_name(self):
        data = {'superQuestion': 'something', 'id': 1}
        body = json.dumps(data)
        response, content = self.client.put(
            path='/api/poll-named/1/', body=body,
            headers={'Content-Type': 'application/json'})

        assert response.status == http.client.OK

        data = json.loads(content.decode('utf8'))

        assert 'superQuestion' in data
        assert data['superQuestion'] == 'something'



# -*- coding: utf-8 -*-
import armet
from armet import exceptions, resources
from .base import BaseResourceTest
import pytest
import json


class TestResource(BaseResourceTest):

    def test_modelless_model_resource(self, connectors):
        with pytest.raises(exceptions.ImproperlyConfigured):
            class Resource(resources.ModelResource):
                pass

    def test_connectorless_resource(self, connectors):
        # Unset configuration.
        old = armet.use.config
        armet.use.config = {}

        with pytest.raises(exceptions.ImproperlyConfigured):
            class Resource(resources.Resource):
                pass

        # Reset the configuration
        armet.use.config = old

    def test_connectorless_abstract_resource(self, connectors):
        # Unset configuration.
        old = armet.use.config
        armet.use.config = {}

        class Resource(resources.Resource):
            class Meta:
                abstract = True

        assert Resource.meta.abstract

        # Reset the configuration
        armet.use.config = old


class TestAllowed(BaseResourceTest):

    def test_http_allowed_methods(self):

        class Resource(resources.ModelResource):

            class Meta:
                abstract = True

                http_allowed_methods = 'GET',

        meta = Resource.meta

        assert 'GET' in meta.http_allowed_methods
        assert 'GET' in meta.http_list_allowed_methods
        assert 'GET' in meta.http_detail_allowed_methods

    def test_http_allowed_methods_to_operations(self):

        class Resource(resources.ModelResource):

            class Meta:
                abstract = True

                http_allowed_methods = 'GET',

        meta = Resource.meta

        assert 'GET' in meta.http_allowed_methods
        assert 'GET' in meta.http_list_allowed_methods
        assert 'GET' in meta.http_detail_allowed_methods

        assert 'read' in meta.list_allowed_operations
        assert 'read' in meta.detail_allowed_operations
        assert 'read' in meta.allowed_operations

    def test_allowed_operations(self):

        class Resource(resources.ModelResource):

            class Meta:
                abstract = True

                allowed_operations = 'read',

        meta = Resource.meta

        assert 'read' in meta.allowed_operations
        assert 'read' in meta.list_allowed_operations
        assert 'read' in meta.detail_allowed_operations

    def test_allowed_operations_to_methods(self):

        class Resource(resources.ModelResource):

            class Meta:
                abstract = True

                allowed_operations = 'read',

        meta = Resource.meta

        assert 'GET' in meta.http_allowed_methods
        assert 'GET' in meta.http_list_allowed_methods
        assert 'GET' in meta.http_detail_allowed_methods

        assert 'read' in meta.list_allowed_operations
        assert 'read' in meta.detail_allowed_operations
        assert 'read' in meta.allowed_operations

    def test_allowed_methods_list(self):

        class Resource(resources.ModelResource):

            class Meta:
                abstract = True

                http_allowed_methods = 'GET',

                http_list_allowed_methods = 'PUT',

        meta = Resource.meta

        assert 'GET' in meta.http_allowed_methods
        assert 'GET' not in meta.http_list_allowed_methods
        assert 'PUT' in meta.http_list_allowed_methods
        assert 'GET' in meta.http_detail_allowed_methods

    def test_allowed_methods_list_to_operations(self):

        class Resource(resources.ModelResource):

            class Meta:
                abstract = True

                http_allowed_methods = 'GET',

                http_list_allowed_methods = 'PUT',

        meta = Resource.meta

        assert 'read' in meta.allowed_operations
        assert 'read' not in meta.list_allowed_operations
        assert 'update' in meta.list_allowed_operations
        assert 'read' in meta.detail_allowed_operations

        assert 'GET' in meta.http_allowed_methods
        assert 'GET' not in meta.http_list_allowed_methods
        assert 'PUT' in meta.http_list_allowed_methods
        assert 'GET' in meta.http_detail_allowed_methods

    def test_allowed_operations_list(self):

        class Resource(resources.ModelResource):

            class Meta:
                abstract = True

                allowed_operations = 'read',

                list_allowed_operations = 'update',

        meta = Resource.meta

        assert 'read' in meta.allowed_operations
        assert 'read' not in meta.list_allowed_operations
        assert 'update' in meta.list_allowed_operations
        assert 'read' in meta.detail_allowed_operations

    def test_allowed_operations_list_to_methods(self):

        class Resource(resources.ModelResource):

            class Meta:
                abstract = True

                allowed_operations = 'read',

                list_allowed_operations = 'update',

        meta = Resource.meta

        assert 'read' in meta.allowed_operations
        assert 'read' not in meta.list_allowed_operations
        assert 'update' in meta.list_allowed_operations
        assert 'read' in meta.detail_allowed_operations

        assert 'GET' in meta.http_allowed_methods
        assert 'GET' not in meta.http_list_allowed_methods
        assert 'PUT' in meta.http_list_allowed_methods
        assert 'GET' in meta.http_detail_allowed_methods


class TestResolution(BaseResourceTest):

    def test_super_direct_resource(self, connectors):
        response, content = self.client.get('/api/direct/')

        assert response.status == 200
        assert content.decode('utf8') == '42'

    def test_super_direct_model_resource(self, connectors):
        response, content = self.client.get('/api/model-direct/')

        assert response.status == 200

        data = json.loads(content.decode('utf8'))

        assert data[0]['question'] == 'Are you an innie or an outie?'

    def test_super_indirect_resource(self, connectors):
        response, content = self.client.get('/api/indirect/')

        assert response.status == 200
        assert content.decode('utf8') == '84'

    def test_super_indirect_model_resource(self, connectors):
        response, content = self.client.get('/api/model-indirect/')

        assert response.status == 200

        data = json.loads(content.decode('utf8'))

        assert data[0]['question'] == 'Are you an innie or an outie?'

    def test_super_twice_indirect_resource(self, connectors):
        response, content = self.client.get('/api/twice-indirect/')

        assert response.status == 200
        assert content.decode('utf8') == '84'

    def test_super_twice_indirect_model_resource(self, connectors):
        response, content = self.client.get('/api/model-twice-indirect/')

        assert response.status == 200

        data = json.loads(content.decode('utf8'))

        assert data[0]['question'] == 'Are you an innie or an outie?'

    def test_super_thrice_indirect_resource(self, connectors):
        response, content = self.client.get('/api/thrice-indirect/')

        assert response.status == 200
        assert content.decode('utf8') == '84'

    def test_super_thrice_indirect_model_resource(self, connectors):
        response, content = self.client.get('/api/model-thrice-indirect/')

        assert response.status == 200

        data = json.loads(content.decode('utf8'))

        assert data[0]['question'] == 'Are you an innie or an outie?'

    def test_mixin_resource(self, connectors):
        response, content = self.client.get('/api/mixin/')

        assert response.status == 200
        assert content.decode('utf8') == 'Hello'



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import six
import operator
from functools import partial
from six.moves import map, reduce
from armet.exceptions import ImproperlyConfigured
from armet.query import parser, Query, QuerySegment, constants
from armet import utils


class ModelResourceOptions(object):

    def __init__(self, meta, name, bases):
        #! SQLAlchemy session used to perform operations on the models.
        self.Session = meta.get('Session')
        if not self.Session:
            raise ImproperlyConfigured(
                'A session factory (via sessionmaker) is required by '
                'the SQLAlchemy model connector.')


def iequal_helper(x, y):
    # String values should use ILIKE queries.
    if isinstance(type(y), six.string_types):
        return x.ilike(y)
    else:
        return operator.eq(x, y)


# Build an operator map to use for sqlalchemy.
OPERATOR_MAP = {
    constants.OPERATOR_EQUAL[0]: operator.eq,
    constants.OPERATOR_IEQUAL[0]: iequal_helper,
    constants.OPERATOR_LT[0]: operator.lt,
    constants.OPERATOR_GT[0]: operator.gt,
    constants.OPERATOR_LTE[0]: operator.le,
    constants.OPERATOR_GTE[0]: operator.ge,
}


def build_segment(model, segment, attr):
    # Get the associated column for the initial path.
    path = segment.path.pop(0)
    col = model.__dict__[path]

    # Resolve the inner-most path segment.
    if segment.path:
        return col.has(build_segment(
            col.property.mapper.class_, segment, attr))

    # Determine the operator.
    op = OPERATOR_MAP[segment.operator]

    # Apply the operator to the values and return the expression
    return reduce(operator.or_,
                  map(partial(op, col),
                      map(attr.try_clean, segment.values)))


def build_clause(query, attributes, model):
    # Iterate through each query segment.
    clause = None
    last = None
    for seg in query.segments:
        # Get the attribute in question.
        attribute = attributes[seg.path[0]]

        # Replace the initial path segment with the expanded
        # attribute path.
        seg.path[0:1] = attribute.path.split('.')

        # Construct the clause from the segment.
        q = build_segment(model, seg, attribute)

        # Combine the segment with the last.
        clause = last.combinator(clause, q) if last is not None else q
        last = seg

    # Return the constructed clause.
    return clause


class ModelResource(object):
    """Specializes the RESTFul model resource protocol for SQLAlchemy.

    @note
        This is not what you derive from to create resources. Import
        ModelResource from `armet.resources` and derive from that.
    """

    def route(self, *args, **kwargs):
        # Establish a session.
        self.session = session = self.meta.Session()

        try:
            # Continue on with the cycle.
            result = utils.super(ModelResource, self).route(*args, **kwargs)

            # Commit the session.
            session.commit()

            # Return the result.
            return result

        except:
            # Something occurred; rollback the session.
            session.rollback()

            # Re-raise the exception.
            raise

        finally:
            # Close the session.
            session.close()

    def filter(self, clause, queryset):
        # Filter the queryset by the passed clause.
        return queryset.filter(clause).distinct()

    def count(self, queryset):
        # Return the count of the queryset.
        return queryset.count()

    def read(self):
        # Initialize the query to the model.
        queryset = self.session.query(self.meta.model)

        query = None
        if self.slug is not None:
            # This is an item-access (eg. GET /<name>/:slug); ignore the
            # query string and generate a query-object based on the slug.
            query = Query(segments=[QuerySegment(
                path=self.meta.slug.path.split('.'),
                operator=constants.OPERATOR_EQUAL[0],
                values=[self.slug])])

        elif self.request.query:
            # This is a list-access; use the query string and construct
            # a query object from it.
            query = parser.parse(self.request.query)

        # Determine if we need to filter the queryset in some way; and if so,
        # filter it.
        clause = None
        if query is not None:
            clause = build_clause(query, self.attributes, self.meta.model)
            queryset = self.filter(clause, queryset)

        # Filter the queryset by asserting authorization.
        queryset = self.meta.authorization.filter(
            self.request.user, 'read', self, queryset)

        # Return the queryset.
        return queryset if self.slug is None else queryset.first()

    def create(self, data):
        # Instantiate a new target.
        target = self.meta.model()

        # Iterate through all attributes and set each one.
        for name, attribute in six.iteritems(self.attributes):
            # Set each one on the target.
            value = data.get(name)
            if value is not None:
                attribute.set(target, value)

        # Ensure the user is authorized to perform this action.
        authz = self.meta.authorization
        if not authz.is_authorized(self.request.user, 'create', self, target):
            authz.unauthorized()

        # Add the target to the session.
        self.session.add(target)
        self.session.flush()

        # Refresh the target object to avoid inconsistencies with storage.
        self.session.expire(target)

        # Return the target.
        return target

    def update(self, target, data):
        # Iterate through all attributes and set each one.
        for name, attribute in six.iteritems(self.attributes):
            # Set each one on the target.
            attribute.set(target, data.get(name))

        # Ensure the user is authorized to perform this action.
        authz = self.meta.authorization
        if not authz.is_authorized(self.request.user, 'update', self, target):
            authz.unauthorized()

        # Flush the target and expire attributes.
        self.session.flush()

        # Refresh the target object to avoid inconsistencies with storage.
        self.session.expire(target)

    def destroy(self):
        # Grab the existing target.
        target = self.read()

        # Ensure the user is authorized to perform this action.
        authz = self.meta.authorization
        if not authz.is_authorized(self.request.user, 'destroy', self, target):
            authz.unauthorized()

        # Remove the object from the session.
        self.session.delete(target)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
from .attribute import Attribute
from .primitive import BooleanAttribute, TextAttribute, IntegerAttribute
from .temporal import TimeAttribute, DateAttribute, DateTimeAttribute
from .uuid import UUIDAttribute
from .timezone import TimezoneAttribute
from .decimal import DecimalAttribute


__all__ = [
    'Attribute',
    'BooleanAttribute',
    'TextAttribute',
    'IntegerAttribute',
    'TimeAttribute',
    'DateAttribute',
    'DateTimeAttribute',
    'UUIDAttribute',
    'TimezoneAttribute',
    'DecimalAttribute'
]



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup, find_packages
import sys
import platform
from imp import load_source


# Required test dependencies.
test_dependencies = [
    # Test runner.
    'pytest',

    # Ensure PEP8 conformance.
    'pytest-pep8',

    # Ensure test coverage.
    'pytest-cov',

    # Benchmarking.
    'pytest-bench',

    # Installs a WSGI application that intercepts requests made to a hostname
    # and port combination for testing.
    'wsgi_intercept >= 0.6.0',

    # HTTP request abstraction layer over httplib.
    'httplib2',

    # The Web framework for perfectionists with deadlines.
    'django',

    # A microframework based on Werkzeug, Jinja2 and good intentions.
    'flask',

    # Bottle is a fast and simple micro-framework for small web applications.
    'bottle',

    # SQLAlchemy is the Python SQL toolkit and Object Relational Mapper
    # that gives application developers the full power and flexibility of SQL.
    'sqlalchemy',
]


if sys.version_info[0] == 2:
    if platform.python_implementation() != 'PyPy':
        # Test dependencies that don't work in PyPy.
        test_dependencies += [
            # gevent is a coroutine-based Python networking library that uses
            # greenlet to provide a high-level synchronous API on top of the
            # libevent event loop.
            'gevent',
        ]


setup(
    name='armet',
    version=load_source('', 'armet/_version.py').__version__,
    description='Clean and modern framework for creating RESTful APIs.',
    author='Concordus Applications',
    author_email='support@concordusapps.com',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Bottle',
        'Framework :: Flask',
        'Framework :: Django',
        # 'Framework :: CherryPy',
        # 'Framework :: Twisted',
        # 'Framework :: Pylons',
        # 'Framework :: Pyramid',
        # 'Framework :: SQLAlchemy',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Database',
        'Topic :: Internet',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    url='http://github.com/armet/python-armet',
    packages=find_packages('.'),
    install_requires=[
        # Python 2 and 3 normalization layer
        'six',

        # For parsing accept and content-type headers
        'python-mimeparse',

        # Ultra fast JSON encoder and decoder for Python.
        'ujson'
    ],
    extras_require={
        'test': test_dependencies
    }
)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import six
import collections
from armet.exceptions import ImproperlyConfigured
from armet.attributes import Attribute
from ..resource.meta import ResourceBase
from . import options


class ManagedResourceBase(ResourceBase):

    options = options.ManagedResourceOptions

    def __new__(cls, name, bases, attrs):
        # Construct the class object.
        self = super(ManagedResourceBase, cls).__new__(cls, name, bases, attrs)

        if not cls._is_resource(name, bases):
            # This is not an actual resource.
            return self

        # Gather declared attributes from ourself and base classes.
        # TODO: We'll likely need a hook here for ORMs
        self.attributes = attributes = collections.OrderedDict()
        for base in bases:
            if getattr(base, 'attributes', None):
                attributes.update(base.attributes)

        for index, attribute in six.iteritems(attrs):
            if isinstance(attribute, Attribute):
                attributes[index] = attribute

        # Ensure all attributes are unique instances and assign names.
        # This is done so that when attributes are resolved; the resolution
        # cache does not clobber base resources (for inherited attributes).
        for attr in attributes:
            attributes[attr] = attributes[attr].clone()
            if not attributes[attr].name:
                attributes[attr].name = attr

        # Resolve the slug reference to an attribute.
        if self.meta.slug not in attributes:
            if not self.meta.abstract:
                raise ImproperlyConfigured(
                    'slug must reference an existing attribute')

        else:
            self.meta.slug = attributes[self.meta.slug]

        # Cache access to the attribute preparation cycle.
        self.preparers = preparers = {}
        for key in attributes:
            prepare = getattr(self, 'prepare_{}'.format(key), None)
            if not prepare:
                prepare = lambda self, obj, value: value
            preparers[key] = prepare

        # Cache access to the attribute clean cycle.
        self.cleaners = cleaners = {}
        for key in attributes:
            clean = getattr(self, 'clean_{}'.format(key), None)
            if not clean:
                clean = lambda self, value: value
            cleaners[key] = clean

        # Return the constructed class object.
        return self
